import static org.junit.Assert.*;
import org.junit.Test;
import org.junit.FixMethodOrder;
import org.junit.runners.MethodSorters;

/**
* A class with unit tests for a MyMap map implementation.
* It tests the MyHashMap class but is designed to be able to
* be extended for testing other MyList implementations.
* For that, subclasses need only override the 
* createEmptyIntList() method.
*
* The tests try to test all methods and corner cases.
* We recommend you tackle them (i.e. try to make them pass) in the 
* order they appear in this file.
*
* @author  Rui Meireles
* @version 1.0
*/
@FixMethodOrder(MethodSorters.NAME_ASCENDING)
public class MyHashMapTest{

    // # integers to use in testing. should be 10 * 2^i to test array at capacity.
    private static final int NINTS = 32768;

    /**
     * A factory method to create an empty map mapping longs to integers.
     * @return an empty map between.
     */
    protected MyHashMap<Long,Integer> createEmptyMap(){
        return new MyHashMap<Long,Integer>();
    }

    /**
     * Tests whether adding map entries is reflected on its size.
     */
    @Test
    public void test1PutAndSize(){
        MyMap<Long,Integer> map = createEmptyMap(); // new empty map
        assertEquals("Initial map size not zero", 0, map.size());

        for (int i=1; i <= NINTS; i++){ // add a map entry at a time, and check map size increases
            Long l = new Long(i);
            Integer v1 = new Integer(i);
            map.put(l, v1);
            assertEquals("Map size doesn't match number of added entries", i, map.size());
            
            Integer v2 = new Integer(i+1);
            map.put(l, v2);
            assertEquals("Map size doesn't match number of added entries", i, map.size());
        }
    }

    /**
     * Tests the behavior of the map's get() and put() methods.
     * Map entries that are added should be able be retrieved correctly.
     * And trying to retrieve non-existing keys should return null.
     */
    @Test
    public void test2PutAndGet(){
        MyMap<Long,Integer> map = createEmptyMap(); // new empty map

        for (int i=1; i <= NINTS; i++){ // add a map entry at a time, and check that element can be retrieved

            Long key = new Long(i);
            Integer ogValue = new Integer(i);

            assertEquals("Map get() doesn't return null when called on non-existing key", null, map.get(key));
            map.put(key, ogValue); // add association to map

            Integer mapValue = map.get(key);
            assertNotEquals("Map get() doesn't return non-null when called on existing key", null, mapValue);

            assertEquals("Map get() doesn't return correct value when called on existing key", ogValue, mapValue);          

            Integer newValue = new Integer(i+1);
            map.put(key, newValue);
            assertEquals("Map get() doesn't return correct value after the value of a pre-existing key is updated", newValue, map.get(key));        
        }

        // go through the map and check all the entries are still there
        for (int i=1; i <= NINTS; i++){ // add a map entry at a time, and check that element can be retrieved

            Long key = new Long(i);
            Integer ogValue = new Integer(i+1);
            assertEquals("Map get() doesn't return correct value when called on existing key", ogValue, map.get(key));          
        }
        
        map = createEmptyMap(); // new empty map
        for (int i=NINTS; i > 0; i--){  // add elements with larger keys first, so they'll change bucket on resize
            Long key = new Long(i);
            Integer value = new Integer(i);
            map.put(key, value); // add association to map
        }
        
        for (int i=NINTS; i > 0; i--){ // check keys are still there after some resizing
            Long key = new Long(i);
            Integer value = new Integer(i);
            assertEquals("Map get() doesn't return correct value when called on existing key", value, map.get(key));          
        }
    }

    /**
     * Tests the behavior of the map's remove method.
     * Assumes correctness of the following methods: put(), size().
     */
    @Test
    public void test3PutAndRemove(){

        MyMap<Long,Integer> map = createEmptyMap(); // new empty map

        for (int i=1; i <= NINTS; i++){ // add a map entry at a time, and check that element can be retrieved
            Long key = new Long(i);
            Integer ogValue = new Integer(i);
            
            assertEquals("Map remove() doesn't return null when called on non-existing key", null, map.remove(key));

            map.put(key, ogValue); // add association to map
            assertEquals("Map remove() doesn't return correct value when called on existing key", ogValue, map.remove(key));
            assertEquals("Map remove() doesn't decrease map size", i-1, map.size());
            
            // add (key, value) back
            map.put(key, ogValue);
            map.put(key, ogValue+1);
            map.put(key, ogValue);
        }

        // go through the map and delete all the entries
        for (int i=NINTS; i > 0; i--){ // add a map entry at a time, and check that element can be retrieved
            Long key = new Long(i);
            Integer ogValue = new Integer(i);
            assertEquals("Map remove() doesn't return correct value when called on existing key", ogValue, map.remove(key));
            assertEquals("Map remove() doesn't decrease map size", i-1, map.size());
        }

    }

    /**
     * Tests that the load factor never goes above the set threshold.
     * 
     * Assumes correctness of the following methods: put(), get().
     */
    @Test
    public void test4LoadFactor(){
        MyHashMap<Long,Integer> map = createEmptyMap(); // new empty map
        final double resizeThreshold = MyHashMap.RESIZE_THRESHOLD;
        
        String emsg = String.format("Map load factor exceeds threshold of %.2f", resizeThreshold);
        
        for (int i=1; i <= NINTS; i++){ // add a map entry at a time, and check that element can be retrieved
            Long key = new Long(i);
            Integer value = new Integer(i);
            map.put(key, value); // add association to map
            Integer mapValue = map.get(key);
            assertEquals("Incorrect value returned for key, either put() or get() aren't working properly.", value, map.get(key));
            assertTrue(emsg, map.loadFactor() <= resizeThreshold);
        }
    }

}
